/*
 * Copyright (C) 2014-2016 Canonical Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 3, as published
 * by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranties of
 * MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of portions of this program with the
 * OpenSSL library under certain conditions as described in each
 * individual source file, and distribute linked combinations
 * including the two.
 * You must obey the GNU General Public License in all respects
 * for all of the code used other than OpenSSL.  If you modify
 * file(s) with this exception, you may extend this exception to your
 * version of the file(s), but you are not obligated to do so.  If you
 * do not wish to do so, delete this exception statement from your
 * version.  If you delete this exception statement from all source
 * files in the program, then also delete it here.
 */

#include <click/application.h>
#include <click/interface.h>
#include "preview.h"
#include <click/qtbridge.h>
#include <click/launcher.h>
#include <click/dbus_constants.h>
#include <click/departments-db.h>
#include <click/utils.h>

#include <boost/algorithm/string/replace.hpp>

#include <unity/UnityExceptions.h>
#include <unity/scopes/CannedQuery.h>
#include <unity/scopes/PreviewReply.h>
#include <unity/scopes/Variant.h>
#include <unity/scopes/VariantBuilder.h>
#include <unity/scopes/ColumnLayout.h>

#include <QCoreApplication>
#include <QDebug>

#include <functional>
#include <iostream>
#include <regex>
#include <sstream>

#include <click/click-i18n.h>

namespace click {

void CachedPreviewWidgets::push(unity::scopes::PreviewWidget const &widget)
{
    widgets.push_back(widget);
    widgets_lookup.insert(widget.id());
}

void CachedPreviewWidgets::push(unity::scopes::PreviewWidgetList const &widgetList)
{
    for (auto const& widget: widgetList)
    {
        push(widget);
    }
}

void CachedPreviewWidgets::flush(unity::scopes::PreviewReplyProxy const& reply)
{
    // A safeguard: if a new widget gets added but missing in the layout, we will get a warning
    // in the logs and layouts will not be registered (single column with all widgets will be used).
    if (widgets.size() != layout.singleColumn.column1.size() ||
            widgets.size() != layout.twoColumns.column1.size() + layout.twoColumns.column2.size())
    {
        qWarning() << "Number of column layouts doesn't match the number of widgets";
    }
    else
    {
        layout.registerLayouts(reply);
    }
    reply->push(widgets);
    widgets.clear();
    widgets_lookup.clear();
}

bool CachedPreviewWidgets::has(std::string const& widget) const
{
    return widgets_lookup.find(widget) != widgets_lookup.end();
}

void WidgetsInColumns::registerLayouts(unity::scopes::PreviewReplyProxy const& reply)
{
    unity::scopes::ColumnLayout layout1col(1);
    layout1col.add_column(singleColumn.column1);

    unity::scopes::ColumnLayout layout2col(2);
    layout2col.add_column(twoColumns.column1);
    layout2col.add_column(twoColumns.column2);

    try
    {
        reply->register_layout({layout1col, layout2col});
    }
    catch (unity::LogicException const& e)
    {
        qWarning() << "Failed to register layout:" << QString::fromStdString(e.what());
    }
}

void WidgetsInColumns::appendToColumn(std::vector<std::string>& column, unity::scopes::PreviewWidgetList const& widgets)
{
    for (auto const& widget: widgets)
    {
        column.push_back(widget.id());
    }
}

DepartmentUpdater::DepartmentUpdater(const std::shared_ptr<click::DepartmentsDb>& depts)
    : depts(depts)
{
}

void DepartmentUpdater::store_department(const PackageDetails& details)
{
    //
    // store package -> department mapping in sqlite db
    if (depts)
    {
        if (!details.department.empty())
        {
            try
            {
                depts->store_package_mapping(details.package.name, details.department);
                qDebug() << "Storing mapping for" << QString::fromStdString(details.package.name) << ":" << QString::fromStdString(details.department);
            }
            catch (const std::exception& e)
            {
                qWarning() << "Failed to store package mapping for package '"
                    << QString::fromStdString(details.package.name)
                    << "', department '" << QString::fromStdString(details.department)
                    << "':" << QString::fromStdString(e.what());
            }
        }
        else
        {
            qWarning() << "Department is empty for package" << QString::fromStdString(details.package.name);
        }
    }
}

// Preview base class

Preview::Preview(const unity::scopes::Result& result,
                 const unity::scopes::ActionMetadata& metadata,
                 std::shared_future<void> const& qt_ready)
    : PreviewQueryBase(result, metadata),
      result(result),
      metadata(metadata),
      qt_ready_(qt_ready)
{
}

Preview::~Preview()
{
}

void Preview::choose_strategy(const QSharedPointer<web::Client> &client,
                              std::shared_ptr<click::DepartmentsDb> depts)
{
    strategy.reset(build_strategy(result, metadata, client, depts));
}

PreviewStrategy* Preview::build_strategy(const unity::scopes::Result &result,
                                         const unity::scopes::ActionMetadata &metadata,
                                         const QSharedPointer<web::Client> &client,
                                         std::shared_ptr<click::DepartmentsDb> depts)
{
    if (result.uri().find("scope://") == 0) {
        return new InstalledScopePreview(result);
    } else {
        return new InstalledPreview(result, metadata, client, depts);
    }

}

void Preview::cancelled()
{
    strategy->cancelled();
}

void Preview::run(const unity::scopes::PreviewReplyProxy &reply)
{
    if (qt_ready_.valid())
        qt_ready_.wait();

    strategy->run(reply);
}

PreviewStrategy::PreviewStrategy(const unity::scopes::Result& result)
    : result(result),
      oa_client("ubuntuone", "ubuntuone", "ubuntuone",
                scopes::OnlineAccountClient::MainLoopSelect::CreateInternalMainLoop)
{
}

PreviewStrategy::PreviewStrategy(const unity::scopes::Result& result,
                 const QSharedPointer<click::web::Client>& client) :
    result(result),
    client(client),
    index(new click::Index(client)),
    oa_client("ubuntuone", "ubuntuone", "ubuntuone",
              scopes::OnlineAccountClient::MainLoopSelect::CreateInternalMainLoop)
{
}

void PreviewStrategy::pushPackagePreviewWidgets(const unity::scopes::PreviewReplyProxy &reply,
                                                const PackageDetails &details,
                                                const scopes::PreviewWidgetList& button_area_widgets)
{
    reply->push(headerWidgets(details));
    reply->push(button_area_widgets);
    reply->push(screenshotsWidgets(details));
    reply->push(descriptionWidgets(details));
}

void PreviewStrategy::pushPackagePreviewWidgets(CachedPreviewWidgets &cache,
                                const PackageDetails& details,
                                const scopes::PreviewWidgetList& button_area_widgets)
{
    cache.push(headerWidgets(details));
    cache.layout.singleColumn.column1.push_back("hdr");
    cache.layout.twoColumns.column1.push_back("hdr");

    cache.push(button_area_widgets);
    cache.layout.appendToColumn(cache.layout.singleColumn.column1, button_area_widgets);
    cache.layout.appendToColumn(cache.layout.twoColumns.column1, button_area_widgets);

    auto const screenshots = screenshotsWidgets(details);
    cache.push(screenshots);
    cache.layout.appendToColumn(cache.layout.singleColumn.column1, screenshots);
    cache.layout.appendToColumn(cache.layout.twoColumns.column1, screenshots);

    auto descr = descriptionWidgets(details);
    if (!descr.empty())
    {
        cache.push(descr);
        cache.layout.appendToColumn(cache.layout.singleColumn.column1, descr);

        // for two-columns we need to split the widgets, summary goes to 1st column, everything else to 2nd
        if (descr.front().id() == "summary")
        {
            descr.pop_front();
            cache.layout.twoColumns.column1.push_back("summary");
        }
        cache.layout.appendToColumn(cache.layout.twoColumns.column2, descr);
    }
}

PreviewStrategy::~PreviewStrategy()
{
}

void PreviewStrategy::cancelled()
{
    index_operation.cancel();
}

scopes::PreviewWidget PreviewStrategy::build_other_metadata(const PackageDetails &details)
{
    scopes::PreviewWidget widget("other_metadata", "table");
    scopes::VariantArray values {
        scopes::Variant{scopes::VariantArray{scopes::Variant{_("Publisher/Creator")}, scopes::Variant{details.publisher}}},
        scopes::Variant{scopes::VariantArray{scopes::Variant{_("Seller")}, scopes::Variant{details.company_name}}},
        scopes::Variant{scopes::VariantArray{scopes::Variant{_("Website")}, scopes::Variant{details.website}}},
        scopes::Variant{scopes::VariantArray{scopes::Variant{_("Contact")}, scopes::Variant{details.support_url}}},
        scopes::Variant{scopes::VariantArray{scopes::Variant{_("License")}, scopes::Variant{details.license}}},
    };
    widget.add_attribute_value("values", scopes::Variant(values));
    return widget;
}

scopes::PreviewWidget PreviewStrategy::build_updates_table(const PackageDetails& details)
{
    scopes::PreviewWidget widget("updates_table", "table");
    widget.add_attribute_value("title", scopes::Variant{_("Updates")});
    scopes::VariantArray values {
        scopes::Variant{scopes::VariantArray{scopes::Variant{_("Version number")}, scopes::Variant{details.version}}},
        scopes::Variant{scopes::VariantArray{scopes::Variant{_("Last updated")}, scopes::Variant{details.last_updated.formatted()}}},
        scopes::Variant{scopes::VariantArray{scopes::Variant{_("First released")}, scopes::Variant{details.date_published.formatted()}}},
        scopes::Variant{scopes::VariantArray{scopes::Variant{_("Size")}, scopes::Variant{click::Formatter::human_readable_filesize(details.binary_filesize)}}},
    };
    widget.add_attribute_value("values", scopes::Variant(values));
    return widget;
}

std::string PreviewStrategy::build_whats_new(const PackageDetails& details)
{
    std::stringstream b;
    b << _("Version") << ": " << details.version << std::endl;
    b << details.changelog;
    return b.str();
}

void PreviewStrategy::run_under_qt(const std::function<void ()> &task)
{
    auto _app = QCoreApplication::instance();
    if (_app != nullptr) {
        qt::core::world::enter_with_task([task]() {
                task();
            });
    } else {
        task();
    }
}

std::string get_string_maybe_null(scopes::Variant variant)
{
    if (variant.is_null()) {
        return "";
    } else {
        return variant.get_string();
    }
}

// Need to use this in a couple places, so it's a static const.
static const std::regex desktop_match("^(.*)\\.desktop$");

// TODO: error handling - once get_details provides errors, we can
// return them from populateDetails and check them in the calling code
// to decide whether to show error widgets. see bug LP: #1289541
void PreviewStrategy::populateDetails(std::function<void(const click::PackageDetails& details)> details_callback,
                                      bool force_cache)
{

    std::string app_name = get_string_maybe_null(result["name"]);

    if (app_name.empty() || std::regex_match(app_name, desktop_match)) {
        click::PackageDetails details;
        qDebug() << "in populateDetails(), app_name is empty";
        details.package.title = result.title();
        details.package.icon_url = result.art();
        details.description = get_string_maybe_null(result["description"]);
        details.main_screenshot_url = get_string_maybe_null(result["main_screenshot"]);
        details_callback(details);
    } else {
        qDebug() << "in populateDetails(), app_name is:" << app_name.c_str();
        // I think this should not be required when we switch the click::Index over
        // to using the Qt bridge. With that, the qt dependency becomes an implementation detail
        // and code using it does not need to worry about threading/event loop topics.
        run_under_qt([this, details_callback, app_name, force_cache]()
            {
                index_operation = index->get_details(app_name, [this, app_name, details_callback, force_cache](PackageDetails details, click::Index::Error error){
                    if(error == click::Index::Error::NoError) {
                        qDebug() << "Got details:" << app_name.c_str();
                        details_callback(details);
                    } else {
                        qDebug() << "Error getting details for:" << app_name.c_str();
                        click::PackageDetails details;
                        details.package.title = result.title();
                        details.package.icon_url = result.art();
                        details.description = get_string_maybe_null(result["description"]);
                        details.main_screenshot_url = get_string_maybe_null(result["main_screenshot"]);
                        details_callback(details);
                    }
                }, force_cache);
            });
    }
}

scopes::PreviewWidgetList PreviewStrategy::screenshotsWidgets(const click::PackageDetails& details)
{
    scopes::PreviewWidgetList widgets;
    bool has_screenshots = !details.main_screenshot_url.empty() || !details.more_screenshots_urls.empty();

    if (has_screenshots)
    {
        scopes::PreviewWidget gallery("screenshots", "gallery");
        scopes::VariantArray arr;

        if (!details.main_screenshot_url.empty())
            arr.push_back(scopes::Variant(details.main_screenshot_url));
        if (!details.more_screenshots_urls.empty())
        {
            for (auto const& s: details.more_screenshots_urls)
            {
                arr.push_back(scopes::Variant(s));
            }
        }

        gallery.add_attribute_value("sources", scopes::Variant(arr));
        widgets.push_back(gallery);
    }
    return widgets;
}

scopes::PreviewWidgetList PreviewStrategy::headerWidgets(const click::PackageDetails& details)
{
    scopes::PreviewWidgetList widgets;

    scopes::PreviewWidget header("hdr", "header");
    header.add_attribute_value("title", scopes::Variant(result.title()));
    if (!details.publisher.empty())
    {
        header.add_attribute_value("subtitle", scopes::Variant(details.publisher));
    }
    if (!details.package.icon_url.empty())
    {
        header.add_attribute_value("mascot", scopes::Variant(details.package.icon_url));
    }

    if (result.contains("price") && result.contains("rating"))
    {
        // Add the price and rating as attributes.
        bool purchased = result["purchased"].get_bool();
        std::string price_area{""};

        if (details.package.price == 0.00f)
        {
            price_area = _("FREE");
        }
        else if (purchased)
        {
            price_area = _("✔ PURCHASED");
        }
        else
        {
            price_area = result["formatted_price"].get_string();
        }
        scopes::VariantBuilder builder;
        builder.add_tuple({
                {"value", scopes::Variant(price_area)},
            });
        builder.add_tuple({
                {"value", scopes::Variant("")},
            });
        builder.add_tuple({
                {"value", result["rating"]},
            });
        builder.add_tuple({
                {"value", scopes::Variant("")},
            });
        header.add_attribute_value("attributes", builder.end());
    }

    widgets.push_back(header);

    qDebug() << "Pushed widgets for package:" << QString::fromStdString(details.package.name);
    return widgets;
}

scopes::PreviewWidgetList PreviewStrategy::descriptionWidgets(const click::PackageDetails& details)
{
    scopes::PreviewWidgetList widgets;
    if (!details.description.empty())
    {
        scopes::PreviewWidget summary("summary", "text");
        summary.add_attribute_value("title", scopes::Variant(_("Info")));
        if (result.contains("description") && !result["description"].get_string().empty())
        {
            summary.add_attribute_value("text", scopes::Variant(result["description"].get_string()));
        }
        else
        {
            summary.add_attribute_value("text", scopes::Variant(details.description));
        }
        widgets.push_back(summary);
    }

    if (!details.download_url.empty())
    {
        widgets.push_back(build_other_metadata(details));
        widgets.push_back(build_updates_table(details));

        scopes::PreviewWidget whats_new("whats_new", "text");
        whats_new.add_attribute_value("title", scopes::Variant(_("What's new")));
        whats_new.add_attribute_value("text", scopes::Variant(build_whats_new(details)));
        widgets.push_back(whats_new);
    }
    return widgets;
}

scopes::PreviewWidgetList PreviewStrategy::errorWidgets(const scopes::Variant& title,
                                                const scopes::Variant& summary,
                                                const scopes::Variant& action_id,
                                                const scopes::Variant& action_label,
                                                const scopes::Variant& uri)
{
    scopes::PreviewWidgetList widgets;

    scopes::PreviewWidget header("hdr", "text");
    header.add_attribute_value("title", title);
    header.add_attribute_value("text", summary);
    widgets.push_back(header);

    scopes::PreviewWidget buttons("buttons", "actions");
    scopes::VariantBuilder builder;
    if (uri.is_null())
    {
        builder.add_tuple({ {"id", action_id}, {"label", action_label} });
    }
    else
    {
        builder.add_tuple({ {"id", action_id}, {"label", action_label}, {"uri", uri} });
    }
    buttons.add_attribute_value("actions", builder.end());
    widgets.push_back(buttons);

    return widgets;
}

// class InstalledPreview

InstalledPreview::InstalledPreview(const unity::scopes::Result& result,
                                   const unity::scopes::ActionMetadata& metadata,
                                   const QSharedPointer<click::web::Client>& client,
                                   const std::shared_ptr<click::DepartmentsDb>& depts)
    : PreviewStrategy(result, client),
      DepartmentUpdater(depts),
      metadata(metadata)
{
}

InstalledPreview::~InstalledPreview()
{
}

void InstalledPreview::run(unity::scopes::PreviewReplyProxy const& reply)
{
    const bool force_cache = (metadata.internet_connectivity() == scopes::QueryMetadata::ConnectivityStatus::Disconnected);
    qDebug() << "preview, force_cache=" << force_cache << ", conn status=" << (int)metadata.internet_connectivity();

    // Get the click manifest.
    Manifest manifest;
    std::string app_name = result["name"].get_string();
    populateDetails([this, reply, manifest, app_name](const PackageDetails &details){
            cachedDetails = details;
            store_department(details);
            pushPackagePreviewWidgets(cachedWidgets, details, createButtons(manifest));
            cachedWidgets.flush(reply);
            reply->finished();
        }, force_cache);
}

scopes::PreviewWidgetList InstalledPreview::createButtons(const Manifest& manifest)
{
    scopes::PreviewWidgetList widgets;
    scopes::PreviewWidget buttons("buttons", "actions");
    scopes::VariantBuilder builder;

    std::string open_label = _("Open");

    auto uri = getApplicationUri(manifest);

    if (!manifest.has_any_apps() && manifest.has_any_scopes()) {
        open_label = _("Search");
    }

    if (!uri.empty())
    {
        builder.add_tuple(
        {
            {"id", scopes::Variant(click::Preview::Actions::OPEN_CLICK)},
            {"label", scopes::Variant(open_label)},
            {"uri", scopes::Variant(uri)}
        });
        qDebug() << "Adding button" << QString::fromStdString(open_label) << "-"
                 << QString::fromStdString(uri);
    }
    if (!uri.empty() || manifest.removable) {
        buttons.add_attribute_value("actions", builder.end());
        widgets.push_back(buttons);
    }
    return widgets;
}

std::string InstalledPreview::getApplicationUri(const Manifest& manifest)
{
    static std::regex appurl_match{"^(application|appid)://[a-zA-Z\\._/-]+$"};

    if (!std::regex_match(result.uri(), appurl_match)) {
        if (manifest.has_any_apps()) {
            std::string uri = "appid://" + manifest.name + "/" +
                manifest.first_app_name + "/current-user-version";
            return uri;
        } else if (manifest.has_any_scopes()) {
            unity::scopes::CannedQuery cq(manifest.first_scope_id);
            auto scope_uri = cq.to_uri();
            qDebug() << "Found uri for scope"
                     << QString::fromStdString(manifest.first_scope_id)
                     << "-" << QString::fromStdString(scope_uri);
            return scope_uri;
        } else {
            qWarning() << "Unable to find app or scope URI for:"
                       << QString::fromStdString(manifest.name);
            return "";
        }
    }
    return result.uri();
}

// class InstalledScopePreview
// this is a temporary fallback preview to get into the Store scope, the proper
// requires 'store' category to be treated special (like 'local') in unity8 shell.

InstalledScopePreview::InstalledScopePreview(const unity::scopes::Result& result)
    : PreviewStrategy(result)
{
}

void InstalledScopePreview::run(unity::scopes::PreviewReplyProxy const& reply)
{
    scopes::PreviewWidget actions("actions", "actions");
    {
        scopes::VariantBuilder builder;
        builder.add_tuple({
                {"id", scopes::Variant("search")},
                {"uri", scopes::Variant(result.uri())},
                {"label", scopes::Variant(_("Search"))}
            });
        actions.add_attribute_value("actions", builder.end());
    }

    reply->push({actions});
}

} // namespace click
